<?php

namespace Drupal\acquia_dam;

/**
 * Create widen specific image url.
 */
final class EmbedCodeUrlBuilder {

  /**
   * Image formats available on the WIDEN end.
   */
  const AVAILABLE_IMAGE_FORMATS = [
    'gif',
    'jpg',
    'png',
    'web',
    'webp',
    'jpeg',
  ];

  /**
   * Converts image effect plugin configuration to query parameters.
   *
   * @param array $effects
   *   Array of effects from a particular style.
   * @param int $origin_width
   *   Width of original image.
   * @param int $origin_height
   *   Height of original image.
   * @param string $uri
   *   Image uri.
   *
   * @return array
   *   Query parameters for embed code URL.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public static function mapImageEffects(array $effects, int $origin_width, int $origin_height, string $uri): array {
    $anchor_mapping = [
      'center-center' => 'c',
      'top' => 'n',
      'right-top' => 'ne',
      'right' => 'e',
      'right-bottom' => 'se',
      'bottom' => 's',
      'left-bottom' => 'sw',
      'left' => 'w',
      'left-top' => 'nw',
    ];
    $values = [];
    $values['format'] = 'web';

    foreach ($effects as $effect) {
      switch ($effect['id']) {
        case 'focal_point_scale_and_crop':
          $values['crop'] = 'yes';
          $values['w'] = $effect['data']['width'];
          $values['h'] = $effect['data']['height'];
          break;

        case 'focal_point_crop':
          if (isset($effect['data']['crop_type'])) {
            $crop_width = $effect['data']['width'];
            $crop_height = $effect['data']['height'];
            // If cropped is bigger than original don't do anything.
            if ($crop_width >= $origin_width || $crop_height >= $origin_height) {
              break;
            }

            /** @var \Drupal\crop\CropInterface $crop */
            if (!$crop = \Drupal::entityTypeManager()
              ->getStorage('crop')
              ->getCrop($uri, $effect['data']['crop_type'])) {
              \Drupal::logger('acquia_dam')->error(sprintf('Cannot find crop entity for %s', $uri));
              break;
            }

            $position = $crop->position();

            $values['crop'] = 'yes';
            $values['w'] = $crop_width;
            $values['h'] = $crop_height;
            $values['a'] =
              self::calculateAnchorPointForWiden($position['x'], $crop_width, $origin_width)
              . ','
              . self::calculateAnchorPointForWiden($position['y'], $crop_height, $origin_height);
          }
          break;

        case 'image_convert':
          if (in_array($effect['data']['extension'], self::AVAILABLE_IMAGE_FORMATS, TRUE)) {
            $values['format'] = $effect['data']['extension'];
          }
          break;

        case 'image_rotate':
          $values['r'] = $effect['data']['degrees'];
          if ($effect['data']['bgcolor'] !== "") {
            $values['color'] = str_replace('#', '', $effect['data']['bgcolor']);
          }
          break;

        case 'image_resize':
          $values['w'] = $effect['data']['width'];
          $values['h'] = $effect['data']['height'];
          break;

        case 'image_crop':
          $values['crop'] = 'yes';
          $values['k'] = $anchor_mapping[$effect['data']['anchor']];
          $values['w'] = $effect['data']['width'];
          $values['h'] = $effect['data']['height'];
          break;

        case 'image_scale':
          if ($effect['data']['width']) {
            $values['w'] = $effect['data']['width'];
          }
          if ($effect['data']['height']) {
            $values['h'] = $effect['data']['height'];
          }
          break;
      }
    }
    return $values;
  }

  /**
   * Anchor point calculation.
   *
   * We use position coordinates to calculate the proper anchor point for WIDEN.
   * Using anchor() method returns position values since focal point does not
   * save crop with and height value.
   *
   * @param int $coordinate
   *   Crop position x or y coordinate.
   * @param int $crop_length
   *   Width or height of the cropped image.
   * @param int $origin_length
   *   Width or height of the original image.
   *
   * @return int
   *   Returns the calculated anchor point.
   */
  protected static function calculateAnchorPointForWiden(int $coordinate, int $crop_length, int $origin_length): int {
    if ($coordinate - $crop_length / 2 <= 0) {
      return 0;
    }
    if ($coordinate + $crop_length / 2 > $origin_length) {
      return $origin_length - $crop_length;
    }

    return $coordinate - $crop_length / 2;
  }

}
