<?php

declare(strict_types=1);

namespace Drupal\acquia_dam\Plugin\media\Source;

use Drupal\acquia_dam\Entity\MediaEmbedsField;
use Drupal\acquia_dam\Entity\MediaSourceField;
use Drupal\acquia_dam\Plugin\Field\FieldType\AssetItem;
use Drupal\Core\File\Exception\FileException;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use Drupal\media\MediaInterface;
use Drupal\media\MediaSourceBase;
use Drupal\media\MediaTypeInterface;
use GuzzleHttp\Exception\TransferException;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Media source for DAM assets.
 *
 * @MediaSource(
 *   id = "acquia_dam_asset",
 *   label = @Translation("Asset"),
 *   description = @Translation("Use an asset from the Acquia DAM"),
 *   allowed_field_types = {"acquia_dam_asset"},
 *   default_thumbnail_filename = "no-thumbnail.png",
 *   deriver = "Drupal\acquia_dam\Plugin\media\Source\AssetDeriver",
 *   asset_search_key = "",
 *   asset_search_value = "",
 * )
 */
final class Asset extends MediaSourceBase {

  /**
   * Asset storage.
   *
   * @var array
   */
  protected $assetData = [];

  /**
   * The DAM client factory.
   *
   * @var \Drupal\acquia_dam\Client\AcquiaDamClientFactory
   */
  private $clientFactory;

  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\Client
   */
  private $httpClient;

  /**
   * The file system.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  private $fileSystem;

  /**
   * The token replacement service.
   *
   * @var \Drupal\Core\Utility\Token
   */
  private $token;

  /**
   * Logger channel interface.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  private $damLoggerChannel;

  /**
   * System date config.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): self {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->clientFactory = $container->get('acquia_dam.client.factory');
    $instance->httpClient = $container->get('http_client');
    $instance->fileSystem = $container->get('file_system');
    $instance->token = $container->get('token');
    $instance->damLoggerChannel = $container->get('logger.channel.acquia_dam');
    $instance->config = $container->get('config.factory')->get('system.date');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getMetadataAttributes(): array {
    $attributes = [
      'created_date' => $this->t('Created date'),
      'filename' => $this->t('File name'),
      'size' => $this->t('Size'),
      'last_update_date' => $this->t('Last updated date'),
      'file_upload_date' => $this->t('File upload date'),
      'expiration_date' => $this->t('Expiration date'),
      'release_date' => $this->t('Release date'),
      'deleted_date' => $this->t('Deleted date'),
    ];
    if ($this->configFactory->get('acquia_dam.settings')->get('allowed_metadata')) {
      return array_merge($attributes, $this->configFactory->get('acquia_dam.settings')->get('allowed_metadata'));
    }
    return $attributes;
  }

  /**
   * {@inheritdoc}
   *
   * Disable PHPMD.CyclomaticComplexity due to the switch statement, which is
   * a pattern used in all the implementations of this method.
   *
   * @SuppressWarnings(PHPMD.CyclomaticComplexity)
   */
  public function getMetadata(MediaInterface $media, $attribute_name) {
    [$asset_id, $version_id] = array_values($this->getSourceFieldValue($media));

    if (empty($asset_id)) {
      return NULL;
    }
    if ($version_id === NULL) {
      $version_id = '';
    }

    $asset = $this->assetData;
    if ($asset === []) {
      try {
        $asset = $this->clientFactory->getSiteClient()->getAsset($asset_id, $version_id);
      }
      catch (\Exception $exception) {
        $this->damLoggerChannel->error(sprintf(
            'Following error occurred while trying to get asset from dam. Asset: %s, error: %s',
            $asset_id,
            $exception->getMessage()
          )
        );
        return NULL;
      }
    }

    switch ($attribute_name) {
      case 'filename':
      case 'default_name':
        return $asset['filename'];

      case 'id':
      case 'version_id':
        return $asset['id'];

      case 'size':
        return $asset['file_properties']['size_in_kbytes'] * 1024;

      case 'thumbnail_uri':
        return $this->getLocalThumbnailUri($asset_id, $version_id, $asset);

      case 'embeds':
        // Check if the field is empty or not if so pull it from the API.
        return $this->getEmbeds($media) ?? $asset['embeds'];

      case 'thumbnails':
        return $asset['thumbnails'];

      case 'image_properties':
        return $asset['file_properties']['image_properties'];

      case 'video_properties':
        return $asset['file_properties']['video_properties'];

      case 'created_date':
      case 'last_update_date':
      case 'file_upload_date':
      case 'deleted_date':
        return $asset[$attribute_name] ? $this->formatDateForDateField($asset[$attribute_name]) : NULL;

      case 'expiration_date':
      case 'release_date':
        return $asset['security'][$attribute_name] ? $this->formatDateForDateField($asset['security'][$attribute_name]) : NULL;

      default:
        // DAM contains multiple metadata type so it requires this check.
        if (array_key_exists($attribute_name, $asset['metadata']['fields'])) {
          return implode(',', $asset['metadata']['fields'][$attribute_name]);
        }
        return NULL;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form = parent::buildConfigurationForm($form, $form_state);
    $form['source_field']['#default_value'] = MediaSourceField::SOURCE_FIELD_NAME;
    $form['source_field']['#disabled'] = TRUE;
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function getSourceFieldValue(MediaInterface $media): array {
    $items = $media->get($this->getSourceFieldName());
    if ($items->isEmpty()) {
      return [
        'asset_id' => '',
        'version_id' => '',
      ];
    }
    $field_item = $items->first();
    assert($field_item instanceof AssetItem);
    // The ::getValue method on FieldItem only returns an array where properties
    // have been initiated with a value. It does not return properties that have
    // no value. Using ::toArray ensures the result has `version_id`, which may
    // be empty when a media item is first saved.
    return $field_item->toArray();
  }

  /**
   * {@inheritdoc}
   */
  public function createSourceField(MediaTypeInterface $type) {
    return $this->getSourceFieldDefinition($type);
  }

  /**
   * {@inheritdoc}
   */
  public function getSourceFieldDefinition(MediaTypeInterface $type) {
    return MediaSourceField::getFieldDefinition('media', $type->id(), $type->label());
  }

  /**
   * {@inheritdoc}
   */
  protected function getSourceFieldName() {
    return MediaSourceField::SOURCE_FIELD_NAME;
  }

  /**
   * Returns the local URI for a resource thumbnail.
   *
   * @param string $asset_id
   *   The asset ID.
   * @param string $version_id
   *   The version ID, an empty string is allowed if version unknown.
   * @param array $asset
   *   The asset.
   *
   * @return string
   *   The local thumbnail URI, or NULL if it could not be downloaded.
   *
   * @see \Drupal\media\Plugin\media\Source\OEmbed::getLocalThumbnailUri
   */
  protected function getLocalThumbnailUri(string $asset_id, string $version_id, array $asset): string {
    if ($version_id === '') {
      // The version ID was not yet set. This happens when the media is saved
      // without a version ID and Media::prepareSave is invoked before preSave
      // where the version is populated. Give a generic version ID for the
      // filename of this thumbnail.
      $version_id = 'default';
    }
    $directory = 'public://acquia_dam_thumbnails/' . $asset_id;
    if (!$this->fileSystem->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS)) {
      return '';
    }
    // If there is an existing local thumbnail for the version, return it.
    $files = $this->fileSystem->scanDirectory($directory, "/^$version_id\..*/");
    if (count($files) > 0) {
      return reset($files)->uri;
    }

    $thumbnail_url = $asset['thumbnails']['160px']['url'];
    try {
      $response = $this->httpClient->request('GET', $thumbnail_url);
      if ($response->getStatusCode() === 200) {
        $local_thumbnail_uri = $directory . DIRECTORY_SEPARATOR . "$version_id.png";
        $this->fileSystem->saveData((string) $response->getBody(), $local_thumbnail_uri, FileSystemInterface::EXISTS_REPLACE);
        return $local_thumbnail_uri;
      }
    }
    catch (TransferException $e) {
      $this->damLoggerChannel->error(sprintf(
          'Unable to download thumbnail at %s: %s %s',
          $thumbnail_url,
          get_class($e),
          $e->getMessage()
        ));
      return '';
    }
    catch (FileException $e) {
      $this->damLoggerChannel->error(sprintf(
        'Unable to download thumbnail at %s: %s %s',
        $thumbnail_url,
        get_class($e),
        $e->getMessage()
      ));
      return '';
    }
    return '';
  }

  /**
   * Helper function that check if the media embeds field as value.
   *
   * @param \Drupal\media\MediaInterface $media
   *   The media object.
   *
   * @return null|string
   *   Returns the embed value if present.
   */
  protected function getEmbeds(MediaInterface $media) {
    if ($media->get(MediaEmbedsField::EMBED_FIELD_NAME)->isEmpty()) {
      return NULL;
    }

    return $media->get(MediaEmbedsField::EMBED_FIELD_NAME)->value;
  }

  /**
   * Formats date coming from DAM to save into storage.
   *
   * @param string $date
   *   Date string coming from API in ISO8601 format.
   *
   * @return string
   *   Date to save into date field value.
   */
  protected function formatDateForDateField(string $date): string {
    $date = \DateTime::createFromFormat(\DateTimeInterface::ISO8601, $date);
    $date->setTimezone(new \DateTimeZone($this->config->get('timezone.default')));
    return $date->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT);
  }

  /**
   * Sets the asset data.
   *
   * @param array $data
   *   The asset data.
   */
  public function setAssetData(array $data) {
    $this->assetData = $data;
  }

}
